/*!
 * @license
 * Copyright Google LLC All Rights Reserved.
 *
 * Use of this source code is governed by an MIT-style license that can be
 * found in the LICENSE file at https://angular.dev/license
 */

import {ComponentFixture, TestBed} from '@angular/core/testing';
import {TableOfContents} from './table-of-contents.component';
import {RouterTestingModule} from '@angular/router/testing';
import {TableOfContentsItem, TableOfContentsLevel} from '../../interfaces';
import {TableOfContentsLoader} from '../../services/table-of-contents-loader.service';
import {TableOfContentsScrollSpy} from '../../services/table-of-contents-scroll-spy.service';
import {WINDOW} from '../../providers';

describe('TableOfContents', () => {
  let component: TableOfContents;
  let fixture: ComponentFixture<TableOfContents>;
  let tableOfContentsLoaderSpy: jasmine.SpyObj<TableOfContentsLoader>;
  let scrollSpy: jasmine.SpyObj<TableOfContentsScrollSpy>;
  const items: TableOfContentsItem[] = [
    {
      title: 'Heading 2',
      top: 0,
      id: 'item-heading-2',
      level: TableOfContentsLevel.H2,
    },
    {
      title: 'First Heading 3',
      top: 100,
      id: 'first-item-heading-3',
      level: TableOfContentsLevel.H3,
    },
    {
      title: 'Second Heading 3',
      top: 200,
      id: 'second-item-heading-3',
      level: TableOfContentsLevel.H3,
    },
  ];
  const fakeWindow = {
    addEventListener: () => {},
    removeEventListener: () => {},
  };

  beforeEach(async () => {
    scrollSpy = jasmine.createSpyObj<TableOfContentsScrollSpy>('TableOfContentsScrollSpy', [
      'startListeningToScroll',
      'activeItemId',
      'scrollbarThumbOnTop',
    ]);
    scrollSpy.startListeningToScroll.and.returnValue();
    scrollSpy.activeItemId.and.returnValue(items[0].id);
    scrollSpy.scrollbarThumbOnTop.and.returnValue(false);
    tableOfContentsLoaderSpy = jasmine.createSpyObj<TableOfContentsLoader>(
      'TableOfContentsLoader',
      ['buildTableOfContent'],
    );
    tableOfContentsLoaderSpy.buildTableOfContent.and.returnValue();
    tableOfContentsLoaderSpy.tableOfContentItems = items;

    await TestBed.configureTestingModule({
      imports: [TableOfContents, RouterTestingModule],
      providers: [
        {
          provide: WINDOW,
          useValue: fakeWindow,
        },
      ],
    }).compileComponents();
    TestBed.overrideProvider(TableOfContentsLoader, {
      useValue: tableOfContentsLoaderSpy,
    });
    TestBed.overrideProvider(TableOfContentsScrollSpy, {
      useValue: scrollSpy,
    });
    fixture = TestBed.createComponent(TableOfContents);
    component = fixture.componentInstance;
    fixture.detectChanges();
  });

  it('should create', () => {
    expect(component).toBeTruthy();
  });

  it('should call scrollToTop when user click on Back to the top button', () => {
    const spy = spyOn(component, 'scrollToTop');

    fixture.detectChanges();

    const button: HTMLButtonElement = fixture.nativeElement.querySelector('button');
    button.click();

    expect(spy).toHaveBeenCalledOnceWith();
  });

  it('should render items when tableOfContentItems has value', () => {
    fixture.detectChanges();

    const renderedItems = fixture.nativeElement.querySelectorAll('li');

    expect(renderedItems.length).toBe(3);
    expect(component.tableOfContentItems().length).toBe(3);
  });

  it('should append level class to element', () => {
    fixture.detectChanges();

    const h2Items = fixture.nativeElement.querySelectorAll('li.docs-toc-item-h2');
    const h3Items = fixture.nativeElement.querySelectorAll('li.docs-toc-item-h3');

    expect(h2Items.length).toBe(1);
    expect(h3Items.length).toBe(2);
  });

  it('should append active class when item is active', () => {
    fixture.detectChanges();

    const activeItem = fixture.nativeElement.querySelector('.adev-faceted-list-item-active');

    expect(activeItem).toBeTruthy();
  });
});
